# -------------------------------------------------------------
# NDN Hydra File Fetcher
# -------------------------------------------------------------
#  @Project: NDN Hydra
#  @Authors: Please check AUTHORS.rst
#  @Source-Code:   https://github.com/tntech-ngin/ndn-hydra
#  @Documentation: https://ndn-hydra.readthedocs.io
#  @Pip-Library:   https://pypi.org/project/ndn-hydra
# -------------------------------------------------------------

import asyncio as aio
import logging
import random
import time
import os
from ndn.app import NDNApp
from ndn.storage import Storage
from ndn_hydra.repo.modules import *
from ndn_hydra.repo.group_messages import *
from ndn_hydra.repo.utils.concurrent_fetcher import concurrent_fetcher


class FileFetcher:
    """
    A class to abstract client-to-node and node-to-node fetching.
    """

    def __init__(self, app: NDNApp, global_view: GlobalView, data_storage: Storage, config: dict) -> None:
        self.app = app
        self.global_view = global_view
        self.data_storage = data_storage
        self.config = config
        self.repo_prefix = config['repo_prefix']
        self.logger = logging.getLogger()
        self.store_func = None  # This function must be initialized to store properly store
        self.fetching = []

    def fetch_file_from_client(self, file_name: str, packets: int, packet_size: int, fetch_path: str):
        if file_name in self.fetching:
            self.logger.info("\nFileFetcher: Already fetching")
            return
        if not self.store_func:
            self.logger.info("\nFileFetcher: No storage function defined")
            return
        self.fetching.append(file_name)
        aio.ensure_future(self._fetch_file_helper(file_name, packets, packet_size, fetch_path))

    def fetch_file_from_node(self, file_name: str, packets: int, packet_size: int):
        if file_name in self.fetching:
            self.logger.info("\nFileFetcher: Already fetching")
            return
        if not self.store_func:
            self.logger.info("\nFileFetcher: No storage function defined")
            return
        self.fetching.append(file_name)
        # Randomly select a node to fetch file from
        file_info = self.global_view.get_file(file_name)
        on_list = file_info["stores"]
        if not on_list:
            self.logger.info("\nFileFetcher: File not in stores")
            return
        active_nodes = set([node['node_name'] for node in self.global_view.get_nodes()])
        on_list = [x for x in on_list if x in active_nodes]
        selected_node = random.choice(on_list)
        if not selected_node:
            return
        # Fetch file from selected node
        fetch_path = f"{self.repo_prefix}/node/{selected_node}/fetch/{file_name}"
        aio.ensure_future(self._fetch_file_helper(file_name, packets, packet_size, fetch_path))

    async def _fetch_file_helper(self, file_name: str, packets: int, packet_size: int, fetch_path: str):
        self.logger.info(f"\n[ACT][FETCH]*  "
                         f"\n\tFile name={file_name};"
                         f"\n\tPackets={packets};"
                         f"\n\tfetch_path={fetch_path}")
        start = time.time()

        async for (_, _, content, data_bytes, key) in concurrent_fetcher(self.app, fetch_path, file_name, 0,
                                                                         packets - 1, aio.Semaphore(15)):
            self.data_storage.put_packet(key, data_bytes)  #TODO: check digest

        end = time.time()
        duration = end - start
        self.logger.info(f"\n[ACT][FETCHED]* "
                         f"\n\tpackets={packets};"
                         f"\n\tduration={duration}")
        self.store_func(file_name)
        try:
            self.fetching.remove(file_name)
        except:
            pass



# -------------------------------------------------------------
# NDN Hydra Global View
# -------------------------------------------------------------
#  @Project: NDN Hydra
#  @Authors: Please check AUTHORS.rst
#  @Source-Code:   https://github.com/tntech-ngin/ndn-hydra
#  @Documentation: https://ndn-hydra.readthedocs.io
#  @Pip-Library:   https://pypi.org/project/ndn-hydra
# -------------------------------------------------------------

import os
import sqlite3
from sqlite3 import Error
from typing import List, Tuple

sql_create_nodes_tables = """
CREATE TABLE IF NOT EXISTS nodes (
    node_name TEXT PRIMARY KEY,
    favor REAL NOT NULL,
    state_vector INTEGER NOT NULL,
    expired INTEGER NOT NULL
);
"""
sql_create_files_tables = """
CREATE TABLE IF NOT EXISTS files (
    file_name TEXT PRIMARY KEY,
    desired_copies INTEGER NOT NULL DEFAULT 3,
    packets INTEGER NOT NULL DEFAULT 1,
    packet_size INTEGER NOT NULL DEFAULT 8800,
    size INTEGER NOT NULL,
    origin_node_name TEXT NOT NULL,
    fetch_path TEXT NOT NULL,
    expiration_time INTEGER NOT NULL
);
"""  # remove origin_node_name, and fetch_path
sql_create_stores_tables = """
CREATE TABLE IF NOT EXISTS stores (
    id INTEGER PRIMARY KEY,
    file_name TEXT NOT NULL,
    node_name NOT NULL
);
"""
sql_create_backups_tables = """
CREATE TABLE IF NOT EXISTS backups (
    id INTEGER PRIMARY KEY,
    file_name TEXT NOT NULL,
    node_name NOT NULL,
    rank INTEGER NOT NULL,
    nonce TEXT NOT NULL
);
"""
sql_create_pending_stores_tables = """
CREATE TABLE IF NOT EXISTS pending_stores (
    id INTEGER PRIMARY KEY,
    file_name TEXT NOT NULL,
    node_name NOT NULL
);
"""


class GlobalView:
    def __init__(self, db: str):
        self.db = os.path.expanduser(db)
        if len(os.path.dirname(self.db)) > 0 and not os.path.exists(os.path.dirname(self.db)):
            try:
                os.makedirs(os.path.dirname(self.db))
            except PermissionError:
                raise PermissionError(f'Could not create database directory: {self.db}') from None
            except FileExistsError:
                pass
        self.__create_tables()

    def __repr__(self):
        self.connection = sqlite3.connect(self.db)
        self.cursor = self.connection.cursor()

        tables = self.get_tables()
        db_repr = []
        for table in tables:
            db_repr.append(f"Table: {table}")
            columns = self.get_columns(table)
            db_repr.append(f"Columns: {', '.join(columns)}")
            rows = self.get_rows(table)
            for row in rows:
                db_repr.append(f"Row: {row}")
            db_repr.append("\n")
        return "\n".join(db_repr)

    def get_tables(self):
        return ["files", "stores", "backups", "nodes", "pending_stores"]

    def get_columns(self, table_name):
        self.cursor.execute(f"PRAGMA table_info({table_name});")
        return [column[1] for column in self.cursor.fetchall()]

    def get_rows(self, table_name):
        self.cursor.execute(f"SELECT * FROM {table_name};")
        return self.cursor.fetchall()

    def close(self):
        self.connection.close()

    def __get_connection(self):
        try:
            return sqlite3.connect(self.db)
        except Error as e:
            print(e)
        return None

    def __execute_sql(self, sql: str):
        result = []
        conn = self.__get_connection()
        if conn is not None:
            try:
                c = conn.cursor()
                c.execute(sql)
                conn.commit()
                result = c.fetchall()
            except Error as e:
                print(e)
            conn.close()
        return result

    def __execute_sql_qmark(self, sql: str, par: Tuple):
        result = []
        conn = self.__get_connection()
        if conn is not None:
            try:
                c = conn.cursor()
                c.execute(sql, par)
                conn.commit()
                result = c.fetchall()
            except Error as e:
                print(e)
            conn.close()
        return result

    def __create_tables(self):
        self.__execute_sql(sql_create_nodes_tables)
        self.__execute_sql(sql_create_files_tables)
        self.__execute_sql(sql_create_stores_tables)
        self.__execute_sql(sql_create_backups_tables)
        self.__execute_sql(sql_create_pending_stores_tables)

    def __rerank_backups(self, file_name: str, node_name: str):
        sql = """
        SELECT DISTINCT file_name, node_name, rank
        FROM backups
        WHERE (file_name = ?) AND (node_name = ?)
        """
        result = self.__execute_sql_qmark(sql, (file_name, node_name))
        rank = -1
        if len(result) == 1:
            rank = result[0][2]
        if rank != -1:
            sql = """
            UPDATE backups
            SET rank = rank - 1
            WHERE (file_name = ?) AND (rank > ?)
            """
            self.__execute_sql_qmark(sql, (file_name, rank))

    def get_node(self, node_name: str):
        sql = """
        SELECT DISTINCT node_name, favor, state_vector, expired
        FROM nodes
        WHERE node_name = ?
        """
        result = self.__execute_sql_qmark(sql, (node_name,))
        if len(result) != 1:
            return None
        return {
            'node_name': result[0][0],
            'favor': result[0][1],
            'state_vector': result[0][2],
            'expired': False if (result[0][3] == 0) else True
        }

    def get_nodes(self, include_expired: bool = False):
        if include_expired:
            sql = """
            SELECT DISTINCT node_name, favor, state_vector, expired
            FROM nodes
            """
        else:
            sql = """
            SELECT DISTINCT node_name, favor, state_vector, expired
            FROM nodes
            WHERE expired = 0
            """
        results = self.__execute_sql(sql)
        nodes = []
        for result in results:
            nodes.append({
                'node_name': result[0],
                'favor': result[1],
                'state_vector': result[2],
                'expired': False if (result[3] == 0) else True
            })
        return nodes

    def get_top_k_nodes(self, k: int):
        # Get top k nodes based on the favor value
        sql = """
        SELECT DISTINCT node_name, favor, state_vector, expired
        FROM nodes
        WHERE expired = 0
        ORDER BY favor DESC
        LIMIT ?
        """
        results = self.__execute_sql_qmark(sql, (k,))
        nodes = []
        for result in results:
            nodes.append({
                'node_name': result[0],
                'favor': result[1],
                'state_vector': result[2]
            })
        return nodes

    def update_node(self, node_name: str, favor: float, state_vector: int):
        sql = """
        INSERT OR REPLACE INTO nodes
            (node_name, favor, state_vector, expired)
        VALUES
            (?, ?, ?, COALESCE((SELECT expired FROM nodes WHERE node_name = ?), 1))
        """
        self.__execute_sql_qmark(sql, (node_name, favor, state_vector, node_name))

    def renew_node(self, node_name: str):
        sql = """
        UPDATE nodes
        SET expired = 0
        WHERE node_name = ?
        """
        self.__execute_sql_qmark(sql, (node_name,))

    def expire_node(self, node_name: str):
        # stores
        sql = """
        DELETE FROM stores WHERE node_name = ?
        """
        self.__execute_sql_qmark(sql, (node_name,))
        # backups
        sql = """
        SELECT file_name, node_name
        FROM backups
        WHERE node_name = ?
        """
        backups = self.__execute_sql_qmark(sql, (node_name,))
        for backup in backups:
            # rerank
            self.__rerank_backups(backup[0], node_name)
        # remove
        sql = """
        DELETE FROM backups WHERE node_name = ?
        """
        self.__execute_sql_qmark(sql, (node_name,))
        # pending_stores
        sql = """
        DELETE FROM pending_stores WHERE node_name = ?
        """
        self.__execute_sql_qmark(sql, (node_name,))
        # expire node
        sql = """
        UPDATE nodes
        SET expired = 1
        WHERE node_name = ?
        """
        self.__execute_sql_qmark(sql, (node_name,))

    def __split_digests(self, digests: bytes, size: int):
        digests_bytes = bytes(digests)
        return [digests_bytes[i:i + size] for i in range(0, len(digests_bytes), size)]

    def get_file(self, file_name: str):
        sql = """
        SELECT DISTINCT
            file_name, desired_copies, packets, size, origin_node_name, fetch_path, packet_size, expiration_time
        FROM files
        WHERE file_name = ?
        """
        result = self.__execute_sql_qmark(sql, (file_name,))
        if len(result) != 1:
            return None
        return {
            'file_name': result[0][0],
            'desired_copies': result[0][1],
            'packets': result[0][2],
            'size': result[0][3],
            'origin_node_name': result[0][4],
            'fetch_path': result[0][5],
            'packet_size': result[0][6],
            'stores': self.get_stores(result[0][0]),
            'backups': self.get_backups(result[0][0]),
            'expiration_time': result[0][7],
        }

    def get_files(self):
        sql = """
        SELECT DISTINCT
            file_name, desired_copies, packets, size, origin_node_name, fetch_path, packet_size, expiration_time
        FROM files
        """
        results = self.__execute_sql(sql)
        files = []
        for result in results:
            files.append({
                'file_name': result[0],
                'desired_copies': result[1],
                'packets': result[2],
                'size': result[3],
                'origin_node_name': result[4],
                'fetch_path': result[5],
                'packet_size': result[6],
                'stores': self.get_stores(result[0]),
                'backups': self.get_backups(result[0]),
                'expiration_time': result[7],
            })
        return files

    def get_underreplicated_files(self):
        files = self.get_files()
        underreplicated_files = []
        for file in files:
            if len(file['stores']) < file['desired_copies']:
                underreplicated_files.append(file)
        return underreplicated_files

    def get_backupable_files(self):
        files = self.get_files()
        backupable_files = []
        for file in files:
            if (len(file['stores']) + len(file['backups'])) < (file['desired_copies'] * 2):
                backupable_files.append(file)
        return backupable_files

    def add_file(self, file_name: str, size: int, origin_node_name: str, fetch_path: str, packet_size: int,
                 packets: int, desired_copies: int, expiration_time: int):
        sql = """
        INSERT OR IGNORE INTO files
            (file_name, desired_copies, packets, size, origin_node_name, fetch_path, packet_size, expiration_time)
        VALUES
            (?, ?, ?, ?, ?, ?, ?, ?)
        """
        self.__execute_sql_qmark(sql, (
        file_name, desired_copies, packets, size, origin_node_name, fetch_path, packet_size, expiration_time))

    def delete_file(self, file_name: str):
        # stores
        sql = """
        DELETE FROM stores WHERE file_name = ?
        """
        self.__execute_sql_qmark(sql, (file_name,))
        # backups
        sql = """
        DELETE FROM backups WHERE file_name = ?
        """
        self.__execute_sql_qmark(sql, (file_name,))
        # backups
        sql = """
        SELECT DISTINCT file_name, node_name
        FROM backups
        WHERE file_name = ?
        """
        backups = self.__execute_sql_qmark(sql, (file_name,))
        for backup in backups:
            # rerank
            self.__rerank_backups(file_name, backup[1])
        # remove
        sql = """
        DELETE FROM backups WHERE file_name = ?
        """
        self.__execute_sql_qmark(sql, (file_name,))
        # pending_stores
        sql = """
        DELETE FROM pending_stores WHERE file_name = ?
        """
        self.__execute_sql_qmark(sql, (file_name,))
        # insertions
        sql = """
        DELETE FROM files WHERE file_name = ?
        """
        self.__execute_sql_qmark(sql, (file_name,))

    def update_file(self, file_name: str, expiration_time: int):
        sql = """
        UPDATE files
        SET expiration_time = ?
        WHERE file_name = ?
        """
        self.__execute_sql_qmark(sql, (expiration_time, file_name))

    def store_file(self, file_name: str, node_name: str):
        # rerank backuped_by
        self.__rerank_backups(file_name, node_name)
        # remove from backuped_by
        sql = """
        DELETE FROM backups WHERE (file_name = ?) AND (node_name = ?)
        """
        self.__execute_sql_qmark(sql, (file_name, node_name))
        # add to stored_by
        sql = """
        INSERT OR IGNORE INTO stores
            (file_name, node_name)
        VALUES
            (?, ?)
        """
        self.__execute_sql_qmark(sql, (file_name, node_name))

    def set_backups(self, file_name: str, backup_list: List[Tuple[str, str]]):
        # remove previous backups
        sql = """
        DELETE FROM backups WHERE (file_name = ?)
        """
        self.__execute_sql_qmark(sql, (file_name,))
        # add backups
        length = len(backup_list)
        for rank in range(length):
            backup = backup_list[rank]
            sql = """
            INSERT OR IGNORE INTO backups
                (file_name, node_name, rank, nonce)
            VALUES
                (?, ?, ?, ?)
            """
            self.__execute_sql_qmark(sql, (file_name, backup[0], rank, backup[1]))

    def add_backup(self, file_name: str, node_name: str, rank: int, nonce: str):
        # delete all backups with larger rank value
        sql = """
        DELETE FROM backups
        WHERE (file_name = ?) AND rank >= ?
        """
        self.__execute_sql_qmark(sql, (file_name, rank))
        # add this backup
        sql = """
        INSERT OR IGNORE INTO backups
            (file_name, node_name, rank, nonce)
        VALUES
            (?, ?, ?, ?)
        """
        self.__execute_sql_qmark(sql, (file_name, node_name, rank, nonce))

    def get_stores(self, file_name: str):
        sql = """
        SELECT DISTINCT file_name, node_name
        FROM stores
        WHERE file_name = ?
        ORDER BY node_name ASC
        """
        results = self.__execute_sql_qmark(sql, (file_name,))
        stores = []
        for result in results:
            stores.append(result[1])
        return stores

    def get_backups(self, file_name: str):
        sql = """
        SELECT DISTINCT file_name, node_name, rank, nonce
        FROM backups
        WHERE file_name = ?
        ORDER BY rank
        """
        results = self.__execute_sql_qmark(sql, (file_name,))
        backups = []
        for result in results:
            backups.append({
                'node_name': result[1],
                'rank': result[2],
                'nonce': result[3]
            })
        return backups

    def get_pending_stores(self, file_name: str):
        sql = """
        SELECT DISTINCT file_name, node_name
        FROM pending_stores
        WHERE file_name = ?
        """
        results = self.__execute_sql_qmark(sql, (file_name,))
        pending_stores = []
        for result in results:
            pending_stores.append(result[1])
        return pending_stores

    def add_pending_store(self, file_name: str, node_name: str):
        sql = """
        INSERT OR IGNORE INTO pending_stores
            (file_name, node_name)
        VALUES
            (?, ?)
        """
        self.__execute_sql_qmark(sql, (file_name, node_name))



# -------------------------------------------------------------
# NDN Hydra Read Handle
# -------------------------------------------------------------
#  @Project: NDN Hydra
#  @Authors: Please check AUTHORS.rst
#  @Source-Code:   https://github.com/tntech-ngin/ndn-hydra
#  @Documentation: https://ndn-hydra.readthedocs.io
#  @Pip-Library:   https://pypi.org/project/ndn-hydra
# -------------------------------------------------------------

import asyncio as aio
import logging
import time
from secrets import choice
from ndn.app import NDNApp
from ndn.encoding import Name, ContentType, Component, parse_data
from ndn.storage import Storage
from ndn_hydra.repo.main.main_loop import MainLoop
from ndn_hydra.repo.modules.global_view import GlobalView
from ndn_hydra.repo.group_messages.update import UpdateMessageTlv
from ndn_hydra.repo.group_messages.message import Message, MessageTypes


class ReadHandle(object):
    """
    ReadHandle processes ordinary interests, and return corresponding data if exists.
    """
    def __init__(self, app: NDNApp, data_storage: Storage, global_view: GlobalView, main_loop: MainLoop, config: dict):
        """
        :param app: NDNApp.
        :param data_storage: Storage.
        :param global_view: Global View.
        :param config: All config Info.
        """
        self.app = app
        self.data_storage = data_storage
        self.global_view = global_view
        self.main_loop = main_loop
        self.node_name = config['node_name']
        self.repo_prefix = config['repo_prefix']
        self.file_expiration = config['file_expiration']

        self.logger = logging.getLogger()

        self.command_comp = "/fetch"
        self.node_comp = "/node"
        # config file needed

        self.listen(Name.from_str(self.repo_prefix + self.command_comp))
        self.listen(Name.from_str(self.repo_prefix + self.node_comp  + self.node_name + self.command_comp))

    def listen(self, prefix):
        """
        This function needs to be called for prefix of all data stored.
        :param prefix: NonStrictName.
        """
        self.app.route(prefix)(self._on_interest)
        self.logger.info(f'\nRead handle: listening to {Name.to_str(prefix)}')

    def unlisten(self, prefix):
        """
        :param name: NonStrictName.
        """
        aio.ensure_future(self.app.unregister(prefix))
        self.logger.info(f'\nRead handle: stop listening to {Name.to_str(prefix)}')

    def _on_interest(self, int_name, int_param, _app_param):
        """
        Repo should not respond to any interest with MustBeFresh flag set.
        Repo will:
        - Reply with data of its own
        - Nack if data can not be found within the repo
        - Reply with a redirect to another node
        Assumptions:
        - A node on the on list will have the file in complete form
        """
        if int_param.must_be_fresh:
            return

        # get rid of the security part if any on the int_name
        file_name = self._get_file_name_from_interest(Name.to_str(int_name[:-1]))
        best_id = self._best_id_for_file(file_name)
        segment_comp = "/" + Component.to_str(int_name[-1])

        if best_id is None:
            if segment_comp == "/seg=0":
                self.logger.info(f'\n[CMD][FETCH]    nacked due to no file')

            # nack due to lack of avaliability
            self.app.put_data(int_name, content=None, content_type=ContentType.NACK)
            self.logger.debug(f"\nRead handle: data not found {Name.to_str(int_name)}")
            return

        if best_id == self.node_name:
            if segment_comp == "/seg=0":
                self.logger.info(f'\n[CMD][FETCH]    serving file')
                self._reset_file_expiration(file_name)

            # serving my own data
            data_bytes = self.data_storage.get_packet(file_name + segment_comp, int_param.can_be_prefix)
            if data_bytes == None:
                return

            self.logger.debug(f'\nRead handle: serve data {Name.to_str(int_name)}')
            _, _, content, _ = parse_data(data_bytes)
            # print("serve"+file_name + segment_comp+"   "+Name.to_str(name))
            final_id = Component.from_number(int(self.global_view.get_file(file_name)["packets"])-1, Component.TYPE_SEGMENT)
            self.app.put_data(int_name, content=content, content_type=ContentType.BLOB, final_block_id=final_id)
        else:
            if segment_comp == "/seg=0":
                self.logger.info(f'\n[CMD][FETCH]    linked to another node')

            # create a link to a node who has the content
            new_name = self.repo_prefix + self.node_comp + best_id + self.command_comp + file_name
            link_content = bytes(new_name.encode())
            final_id = Component.from_number(int(self.global_view.get_file(file_name)["packets"])-1, Component.TYPE_SEGMENT)
            self.app.put_data(int_name, content=link_content, content_type=ContentType.LINK, final_block_id=final_id)

    def _get_file_name_from_interest(self, int_name):
        file_name = int_name[len(self.repo_prefix):]
        if file_name[0:len(self.node_comp)] == self.node_comp:
            return file_name[(len(self.node_comp)+len(self.node_name)+len(self.command_comp)):]
        else:
            return file_name[(len(self.command_comp)):]

    def _best_id_for_file(self, file_name: str):
        file_info = self.global_view.get_file(file_name)
        if not file_name or not file_info:
            return None
        active_nodes = set( [x['node_name'] for x in self.global_view.get_nodes()] )
        on_list = file_info["stores"]
        if not on_list:
            return None
        if self.node_name in on_list:
            return self.node_name
        else:
            on_list = [x for x in on_list if x in active_nodes]
            return choice(on_list)

    def _reset_file_expiration(self, file_name):
        if self.file_expiration == 0:  # no need to reset if file_expiration in config is set to 0
            return
        expiration_time = int(time.time() + (self.file_expiration * 60 * 60))  # convert hours to seconds

        # update tlv
        favor = self.global_view.get_node(self.node_name)['favor']
        update_message = UpdateMessageTlv()
        update_message.node_name = self.node_name.encode()
        update_message.favor = str(favor).encode()
        update_message.file_name = file_name
        update_message.expiration_time = expiration_time
        # update msg
        message = Message()
        message.type = MessageTypes.UPDATE
        message.value = update_message.encode()

        # publish
        self.global_view.update_file(file_name, expiration_time)
        self.main_loop.svs.publishData(message.encode())


import os
import yaml


def read_config_file():
    default_config = {}

    script_dir = os.path.dirname(__file__)
    config_path = os.path.join(script_dir, '../config.yaml')
    try:
        with open(config_path, "r") as yamlfile:
            default_config = yaml.safe_load(yamlfile)

    except Exception as e:
        print(f"\n > An error occurred while reading config file: {e}")

    return default_config



"""
Read the remaining space from the node
"""

import shutil


def convert_bytes(bytes_value):
    for unit in ["B", "KB", "MB", "GB", "TB"]:
        if bytes_value < 1024:
            return f"{bytes_value:.2f} {unit}"
        bytes_value /= 1024


def get_remaining_space(current_node):
    disk_usage = shutil.disk_usage(current_node)
    free_space = disk_usage.free / 1024 / 1024 / 1024

    return free_space


import asyncio as aio


# Responsible for removing data from the data storage
def remove_file(data_storage, file):
    keys = [file['file_name'] + f'/seg={seq}' for seq in range(file['packets'])]
    aio.get_event_loop().run_in_executor(None, data_storage.remove_packets, keys)



# -------------------------------------------------------------
# NDN Hydra Insert Client
# -------------------------------------------------------------
#  @Project: NDN Hydra
#  @Authors: Please check AUTHORS.rst
#  @Source-Code:   https://github.com/tntech-ngin/ndn-hydra
#  @Documentation: https://ndn-hydra.readthedocs.io
#  @Pip-Library:   https://pypi.org/project/ndn-hydra
# -------------------------------------------------------------

import logging
import time
from hashlib import blake2b
from ndn.app import NDNApp
from ndn.encoding import Name, Component, FormalName
from ndn_hydra.repo.protocol.base_models import InsertCommand, File
from ndn_hydra.repo.utils.pubsub import PubSub
from ndn_hydra.client.functions.query import HydraQueryClient

SEGMENT_SIZE = 8192


class HydraInsertClient(object):
    def __init__(self, app: NDNApp, client_prefix: FormalName, repo_prefix: FormalName) -> None:
        """
        This client inserts data packets from the remote repo.
        :param app: NDNApp.
        :param client_prefix: NonStrictName. Routable name to client.
        :param repo_prefix: NonStrictName. Routable name to remote repo.
        """
        self.app = app
        self.client_prefix = client_prefix
        self.repo_prefix = repo_prefix
        self.pb = PubSub(self.app, self.client_prefix)
        self.packets = []

    async def insert_file(self, file_name: FormalName, path: str) -> bool:
        """
        Insert a file associated with a file name to the remote repo
        """
        size, seg_cnt = 0, 0
        fetch_file_prefix = self.client_prefix + [Component.from_str("upload")] + file_name

        # Check if the file already exists
        query_client = HydraQueryClient(self.app, self.client_prefix, self.repo_prefix)
        query = [Component.from_str("file")] + file_name
        query_result = await query_client.send_query(query)
        target_file_name = Name.to_str(query)[5:]

        if query_result == target_file_name:
            print('File already exists, aborted insertion.')
            return False

        tic = time.perf_counter()
        with open(path, "rb") as f:
            data = f.read()
            size = len(data)
            seg_cnt = (len(data) + SEGMENT_SIZE - 1) // SEGMENT_SIZE
            self.packets = [self.app.prepare_data(fetch_file_prefix + [Component.from_segment(i)],
                                                  data[i * SEGMENT_SIZE:(i + 1) * SEGMENT_SIZE],
                                                  freshness_period=10000,
                                                  final_block_id=Component.from_segment(seg_cnt - 1))
                            for i in range(seg_cnt)]

        print(f'\nCreated {seg_cnt} chunks under name {Name.to_str(fetch_file_prefix)}')

        def on_interest(int_name, _int_param, _app_param):
            seg_no = Component.to_number(int_name[-1]) if Component.get_type(
                int_name[-1]) == Component.TYPE_SEGMENT else 0
            if seg_no < seg_cnt:
                self.app.put_raw_packet(self.packets[seg_no])
            if seg_no == (seg_cnt - 1):
                toc = time.perf_counter()
                print(f"The publication is complete! - total time (with disk): {toc - tic:0.4f} secs")

        self.app.route(fetch_file_prefix)(on_interest)

        file = File()
        file.file_name = file_name
        file.packets = seg_cnt
        file.packet_size = SEGMENT_SIZE
        file.size = size
        cmd = InsertCommand()
        cmd.file = file
        cmd.fetch_path = fetch_file_prefix
        cmd_bytes = cmd.encode()

        # publish msg to repo's insert topic
        await self.pb.wait_for_ready()
        is_success = await self.pb.publish(self.repo_prefix + ['insert'], cmd_bytes)
        if is_success:
            logging.debug('\nPublished an insert msg and was acknowledged by a subscriber')
        else:
            logging.debug('\nPublished an insert msg but was not acknowledged by a subscriber')
        return is_success
